use std::env;
use std::fs::File;
use std::io::{BufRead, BufReader};
use std::path::{Path, PathBuf};



/// The default name of the resources directory.
const DEFAULT_RESOURCES_DIR: &'static str = "resources";

/// The default name to be used for the resource manifest file.
const DEFAULT_MANIFEST_FILENAME: &'static str = "resources.msnr";



/// Defines a resource to be compiled.
struct Section
{
   /// A unique name used to reference the
   /// resource defined in this Section.
   pub name: String,

   /// A hint for what kind of loader to
   /// use for this resource.
   pub type_hint: String,

   /// The path to the resources src information.
   pub src_path: String,

   /// The path to place the compiled resource at.
   pub dst_path: String,

   /// The compiler to invoke to compile the
   /// resource defined in this Section.
   pub compiler: String,

   /// The arguments to pass to the compiler.
   pub arguments: String
}

/// Processes the "resources.msnr" file and turns it
/// into a "resources.msnl" file.
pub struct Processor
{
   /// The name of the resource manifest file.
   manifest_filename: String,

   /// The path to the resources directory.
   resources_dir: PathBuf,

   /// The path to the output directory.
   output_dir: PathBuf,

   /// The amount of tasks to use when processing the
   /// resources from the manifest file.
   task_count: u64,

   /// The resources sections from the "resources.msnr" file.
   sections: Vec<Section>
}



impl Section
{
   /// Create a new Section with default values.
   pub fn new() -> Section
   {
      Section
      {
         name: String::new(),
         type_hint: String::new(),
         src_path: String::new(),
         dst_path: String::new(),
         compiler: String::new(),
         arguments: String::new()
      }
   }
}

impl Processor
{
   /// Create a new resource manifest Processor.
   pub fn new() -> Processor
   {
      let out_dir: PathBuf;

      // Try to just get the current directory.
      match env::current_dir()
      {
         Ok(dir) =>
         {
            out_dir = PathBuf::from(dir);
         }

         Err(error) =>
         {
            warn!("{}", error);
            out_dir = PathBuf::new();
         }
      }

      Processor
      {
         manifest_filename: String::from(DEFAULT_MANIFEST_FILENAME),
         resources_dir: PathBuf::from(DEFAULT_RESOURCES_DIR),
         output_dir: out_dir,
         task_count: 1u64,
         sections: Vec::new()
      }
   }

   /// Change the directory to load the resource manifest file from.
   /// This is also the directory used as the root when looking for
   /// resources.
   ///
   /// Defaults to "".
   pub fn set_resources_dir(&mut self, dir: &Path)
   {
      self.resources_dir = PathBuf::from(dir);
      println!("Resource dir: {}", self.resources_dir.to_str().unwrap());
   }

   /// Change the directory to place the compiled resources into.
   /// This is also the directory where the loaded manifest file will
   /// be placed.
   ///
   /// Defaults to "resources".
   pub fn set_output_dir(&mut self, dir: &Path)
   {
      self.output_dir = PathBuf::from(dir);
      println!("Output dir: {}", self.output_dir.to_str().unwrap());
   }

   /// Sets the amount of tasks to have live to compile the resources.
   ///
   /// Each resource will be compiled on a seperate blocking thread.
   pub fn set_task_count(&mut self, count: u64)
   {
      self.task_count = count;
      println!("Task count: {}", self.task_count);
   }

   /// Process the resource manifest file.
   pub fn process_manifest_file(&mut self)
   {
      // Read the manifest file from the resources directory.
      self.read_manifest_file();

      // Compile all the resources from the manifest file.
      self.compile_resources();
   }

   /// Read the manifest file.
   ///
   /// Reading the manifest file can be a blocking
   /// operation since there is nothing else to do until
   /// the contents of this file have been read.
   fn read_manifest_file(&mut self)
   {
      let file: File;
      let mut continue_reading: bool;
      let mut line: String;
      let mut filepath: PathBuf;
      let mut reader: BufReader<File>;
      let mut section_buffer: Vec<String>;

      // Create the filepath for the resource manifest.
      filepath = self.resources_dir.clone();
      filepath.push(self.manifest_filename.clone());

      // Open the manifest file for reading.
      match File::open(filepath)
      {
         Ok(opened_file) =>
         {
            file = opened_file
         }

         Err(error) =>
         {
            error!("{} -- {:?}  {:?}", error,
                   self.resources_dir, self.manifest_filename);
         }
      }

      // Create a reader to read all of the sections of
      // the manifest file.
      reader = BufReader::new(file);

      // Create a buffer to hold the current
      // sections information.
      section_buffer = Vec::new();

      // Read all of the file line by line.
      continue_reading = true;
      while continue_reading == true
      {
         // Read a line from the file.
         line = String::new();
         match reader.read_line(&mut line)
         {
            Ok(num_bytes) =>
            {
               if num_bytes > 0
               {
                  if line == "\n"
                  {
                     // This is an empty line so if there are any
                     // lines in the section buffer then they need
                     // to be read as a section. If not then there
                     // is nothing to do.
                     self.read_manifest_section(&section_buffer);
                     section_buffer = Vec::new();
                  }
                  else
                  {
                     // Add this line to the section buffer.
                     section_buffer.push(line.clone());
                  }
               }
               else
               {
                  // It looks like there is nothing more to be read.
                  continue_reading = false;
               }
            }

            Err(error) =>
            {
               error!("{}", error);
            }
         }
      }

      // Check to make sure that anythin left in
      // the section buffer is handled.
      self.read_manifest_section(&section_buffer);
   }

   /// Read a single manifest section from a given buffer.
   fn read_manifest_section(&mut self, buffer: &Vec<String>)
   {
      let mut header: &str;
      let mut value: &str;
      let mut section: Section;
      let mut values: Vec<&str>;

      // Only do something if there are some
      // lines in the buffer to process.
      if buffer.len() > 0
      {
         // Create a new Section and fill out its information
         // by parsing the data in the buffer.
         section = Section::new();
         for line in buffer.iter()
         {
            // Split each line in the buffer by the '=' character.
            // This should be parsed into 2 lines then.
            values = line.split("=").collect();
            if values.len() == 2
            {
               // Get the header and value to set
               // the Sections values.
               header = values[0].trim();
               value = values[1].trim();
               if header == "name"
               {
                  section.name = String::from(value);
               }
               else if header == "type"
               {
                  section.type_hint = String::from(value);
               }
               else if header == "src"
               {
                  section.src_path = String::from(value);
               }
               else if header == "dst"
               {
                  section.dst_path = String::from(value);
               }
               else if header == "compiler"
               {
                  section.compiler = String::from(value);
               }
               else if header == "args"
               {
                  section.arguments = String::from(value);
               }
            }
            else
            {
               warn!("Detected a name value entry that is incorrect.\n{}",
                     line.trim());
            }
         }

         // Add the newly created section to the list of sections.
         self.sections.push(section);
      }
   }

   /// Compile all the resources in the Section list.
   fn compile_resources(&self)
   {
   }
}
